import errno
import json
import logging
import os
import re
import shutil
import sys
import stat
import time
from subprocess import CalledProcessError

from .cloud import CloudCli, AWSCli, AzureCli, GCloudCli
from .terraform import TerraformCli
from .ansible import AnsibleCli
from .action import ActionManager as AM
from .specifications import default_spec, merge_user_spec
from .spec.reference_architecture import ReferenceArchitectureSpec
from .system import exec_shell


class ProjectError(Exception):
    pass


class Project:

    projects_root_path = os.getenv(
        'EDB_DEPLOY_DIR',
        os.path.join(os.path.expanduser("~"), ".edb-pot")
    )
    # Path that should contain 3rd party tools binaries when they are installed
    # by the prerequisites installation script.
    cloud_tools_bin_path = os.path.join(
        os.path.expanduser("~"),
        '.edb-cloud-tools',
        'bin'
    )
    terraform_share_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        'data',
        'terraform'
    )
    ansible_share_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        'data',
        'ansible'
    )
    terraform_templates = ['variables.tf.template', 'tags.tf.template']
    ansible_collection_name = 'edb_devops.edb_postgres:3.2.0'
    aws_collection_name = 'community.aws:1.4.0'

    def __init__(self, cloud, name):
        self.name = name
        self.cloud = cloud
        self.project_path = os.path.join(
            self.projects_root_path,
            self.cloud,
            self.name
        )
        self.terraform_path = os.path.join(
            self.terraform_share_path,
            self.cloud
        )
        self.terraform_plugin_cache_path = os.path.join(
            self.projects_root_path,
            '.terraform_plugin_cache'
        )
        self.ssh_priv_key = os.path.join(self.project_path, 'ssh_priv_key')
        self.ssh_pub_key = os.path.join(self.project_path, 'ssh_pub_key')

        self.terraform_vars = None
        self.terraform_vars_file = os.path.join(
            self.project_path,
            'terraform_vars.json'
        )
        self.ansible_vars = None
        self.ansible_vars_file = os.path.join(
            self.project_path,
            'ansible_vars.json'
        )
        self.reference_architecture = ReferenceArchitectureSpec
        self.log_file = os.path.join(
            self.projects_root_path, "log", self.cloud, "%s.log" % self.name
        )
        self.ansible_playbook = os.path.join(self.project_path, 'playbook.yml')
        self.ansible_inventory = os.path.join(
            self.project_path,
            'inventory.yml'
        )

        self.ansible_pot_role = os.path.join(self.ansible_share_path, 'roles')

        self.state_file = os.path.join(self.project_path, 'state.json')
        self.custom_ssh_keys = {}

    def check_versions(self):
         # Check Ansible version
         ansible = AnsibleCli('dummy', bin_path=self.cloud_tools_bin_path)
         ansible.check_version()
         # Check Terraform version
         terraform = TerraformCli('dummy', 'dummy',
                                  bin_path=self.cloud_tools_bin_path)
         terraform.check_version()
         # Check cloud vendor CLI/SDK version
         cloud_cli = CloudCli(self.cloud, bin_path=self.cloud_tools_bin_path)
         cloud_cli.check_version()

    def create_log_dir(self):
        try:
            os.makedirs(os.path.dirname(self.log_file))
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise ProjectError(str(e))
        except Exception as e:
            raise ProjectError(str(e))

    @staticmethod
    def create_root_log_dir():
        try:
            os.makedirs(os.path.join(Project.projects_root_path, "log"))
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise ProjectError(str(e))
        except Exception as e:
            raise ProjectError(str(e))

    def exists(self):
        return os.path.exists(self.project_path)

    def init_state(self):
        with open(self.state_file, 'w') as f:
            f.write(json.dumps(dict()))

    def update_state(self, component, state):
        try:
            if not os.path.exists(self.state_file):
                self.init_state()

            with open(self.state_file, 'r') as f:
                states = json.loads(f.read())
            # Update component's state
            states[component] = state
            # Save the state file
            with open(self.state_file, 'w') as f:
                f.write(json.dumps(states))
        except Exception as e:
            msg = "Unable to update the state file %s" % self.state_file
            logging.error(msg)
            logging.exception(str(e))
            raise ProjectError(msg)

    def load_states(self):
        try:
            with open(self.state_file, 'r') as f:
                return json.loads(f.read())
        except Exception as e:
            msg = "Unable to read the state file %s" % self.state_file
            logging.error(msg)
            logging.exception(str(e))
            raise ProjectError(msg)

    def create(self):
        # Copy terraform code
        with AM("Copying Terraform code from into %s" % self.project_path):
            try:
                shutil.copytree(self.terraform_path, self.project_path)
            except Exception as e:
                raise ProjectError(str(e))

        with AM("Initialzing state file"):
            self.init_state()

        # Create Terraform plugin cache
        with AM(
            "Creating Terraform plugin cache dir. %s"
            % self.terraform_plugin_cache_path
        ):
            try:
                os.makedirs(self.terraform_plugin_cache_path)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise ProjectError(str(e))
            except Exception as e:
                raise ProjectError(str(e))

    def configure(self, env):
        # Copy the PoT role in ansible project directory
        with AM("Copying PoT role code from into %s" % os.path.join(self.project_path, "roles")):
            try:
                shutil.copytree(self.ansible_pot_role, os.path.join(self.project_path, "roles"))
            except Exception as e:
                raise ProjectError(str(e))

        with AM("Loading Cloud specifications"):
            if getattr(env, 'spec_file', False):
                user_spec = self._load_spec_file(env.spec_file.name)
                env.cloud_spec = merge_user_spec(env.cloud, user_spec)
            else:
                env.cloud_spec = default_spec(env.cloud)

            logging.debug("env.cloud_specs=%s", env.cloud_spec)

        with AM("Creating ssh keys for project"):
            ssh_user = env.cloud_spec['available_os'][env.operating_system]['ssh_user']
            self.ssh_key_gen(ssh_user, False)

        with AM("Creating customer ssh keys for project"):
            self.ssh_key_gen(self.name, True)

        # Transform templates
        for template in self.terraform_templates:
            template_path = os.path.join(self.project_path, template)
            if not os.path.exists(template_path):
                continue
            dest_path = os.path.join(
                self.project_path,
                os.path.splitext(template)[0]
            )
            with AM(
                "Generating file %s from template %s" % (dest_path, template)
            ):
                with open(template_path, 'r') as f:
                    with open(dest_path, 'w') as d:
                        for l in f.readlines():
                            d.write(l.replace("%PROJECT_NAME%", self.name))
                os.unlink(template_path)

        # Build the vars files for Terraform and Ansible
        with AM("Building Terraform vars file %s" % self.terraform_vars_file):
            self._build_terraform_vars(env)
            logging.debug("terraform_vars=%s", self.terraform_vars)
            # Save Terraform vars
            self._save_terraform_vars()

        with AM("Building Ansible vars file %s" % self.ansible_vars_file):
            self._build_ansible_vars(env)
            logging.debug("ansible_vars=%s", self.ansible_vars)
            # Save Ansible vars
            self._save_ansible_vars()

        with AM("Copying Ansible playbook file %s" % self.ansible_playbook):
            shutil.copy(
                os.path.join(
                    self.ansible_share_path,
                    "%s.yml" % self.ansible_vars['reference_architecture']
                ),
                self.ansible_playbook
            )


        # Instanciate a new CloudCli
        cloud_cli = CloudCli(env.cloud, bin_path=self.cloud_tools_bin_path)

        # Build a list of instance_type accordingly to the specs
        instance_types = []
        node_types = ['postgres_server', 'pem_server', 'barman_server',
                      'pooler_server']
        for node_type in node_types:
            node = self.terraform_vars.get(node_type)
            if not node:
                continue
            if node['instance_type'] not in instance_types:
                instance_types.append(node['instance_type'])

        # AWS - Check instance type and image availability
        if env.cloud == 'aws' and not self.terraform_vars['aws_ami_id']:
            for instance_type in instance_types:
                with AM(
                    "Checking instance type %s availability in %s"
                    % (instance_type, env.aws_region)
                ):
                    cloud_cli.check_instance_type_availability(
                        instance_type, env.aws_region
                    )

            # Check availability of image in target region and get its ID
            with AM(
                "Checking image '%s' availability in %s"
                % (self.terraform_vars['aws_image'], env.aws_region)
            ):
                aws_ami_id = cloud_cli.cli.get_image_id(
                    self.terraform_vars['aws_image'], env.aws_region
                )
                if not aws_ami_id:
                    raise ProjectError(
                        "Unable to get Image Id for image %s in region %s"
                        % (self.terraform_vars['aws_image'], env.aws_region)
                    )
            with AM("Updating Terraform vars with the AMI id %s" % aws_ami_id):
                # Useless variable for Terraform
                del(self.terraform_vars['aws_image'])
                self.terraform_vars['aws_ami_id'] = aws_ami_id
                self._save_terraform_vars()

        # Azure - Check instance type and image availability
        if env.cloud == 'azure':
            for instance_type in instance_types:
                with AM(
                    "Checking instance type %s availability in %s"
                    % (instance_type, env.azure_region)
                ):
                    cloud_cli.check_instance_type_availability(
                        instance_type, env.azure_region
                    )
            # Check availability of image in target region
            with AM(
                "Checking image %s:%s:%s availability in %s"
                % (
                    self.terraform_vars['azure_publisher'],
                    self.terraform_vars['azure_offer'],
                    self.terraform_vars['azure_sku'],
                    env.azure_region
                  )
            ):
                cloud_cli.cli.check_image_availability(
                    self.terraform_vars['azure_publisher'],
                    self.terraform_vars['azure_offer'],
                    self.terraform_vars['azure_sku'],
                    env.azure_region
                )

        # GCloud - Check instance type and image availability
        if env.cloud == 'gcloud':
            for instance_type in instance_types:
                with AM(
                    "Checking instance type %s availability in %s"
                    % (instance_type, env.gcloud_region)
                ):
                    cloud_cli.check_instance_type_availability(
                        instance_type, env.gcloud_region
                    )
            # Check availability of the image
            with AM(
                "Checking image %s availability"
                % self.terraform_vars['gcloud_image']
            ):
                cloud_cli.cli.check_image_availability(
                    self.terraform_vars['gcloud_image']
                )

    def _load_spec_file(self, spec_file_path):
        try:
            with open(spec_file_path) as json_file:
                data = json.loads(json_file.read())
            return data
        except json.decoder.JSONDecodeError as e:
            msg = "Unable to load the JSON spec. file %s" % spec_file_path
            logging.error(msg)
            logging.error(str(e))
            raise ProjectError(msg)

    def _load_terraform_vars(self):
        try:
            with open(self.terraform_vars_file) as json_file:
                self.terraform_vars = json.loads(json_file.read())
        except json.decoder.JSONDecodeError as e:
            msg = ("Unable to load the Terraform vars file %s"
                   % self.terraform_vars_file)
            logging.error(msg)
            logging.error(str(e))
            raise ProjectError(msg)

    def _save_terraform_vars(self):
        try:
            with open(self.terraform_vars_file, "w") as json_file:
                json_file.write(json.dumps(self.terraform_vars, indent=2))
        except Exception as e:
            msg = ("Unable to save the Terraform vars file %s"
                    % self.terraform_vars_file)
            logging.error(msg)
            logging.error(str(e))
            raise ProjectError(msg)

    def _load_ansible_vars(self):
        try:
            with open(self.ansible_vars_file) as json_file:
                self.ansible_vars = json.loads(json_file.read())
        except json.decoder.JSONDecodeError as e:
            msg = ("Unable to load the Ansible vars file %s"
                   % self.ansible_vars_file)
            logging.error(msg)
            logging.error(str(e))
            raise ProjectError(msg)

    def _save_ansible_vars(self):
        try:
            with open(self.ansible_vars_file, "w") as json_file:
                json_file.write(json.dumps(self.ansible_vars, indent=2))
        except Exception as e:
            msg = ("Unable to save the Ansible vars file %s"
                    % self.ansible_vars_file)
            logging.error(msg)
            logging.error(str(e))
            raise ProjectError(msg)

    def _build_terraform_vars(self, env):
        ra = self.reference_architecture[env.reference_architecture]
        pg_spec = env.cloud_spec['postgres_server']
        os_spec = env.cloud_spec['available_os'][env.operating_system]

        self.terraform_vars = dict(
            cluster_name=self.name,
            replication_type=ra['replication_type'],
            ssh_user=os_spec['ssh_user'],
            ssh_priv_key=self.custom_ssh_keys[os_spec['ssh_user']]['ssh_priv_key'],
            ssh_pub_key=self.custom_ssh_keys[os_spec['ssh_user']]['ssh_pub_key'],
            barman=ra['barman'],
            pooler_local=ra['pooler_local'],
            pooler_type=ra['pooler_type']
        )

        # AWS case
        if env.cloud == 'aws':
            self.terraform_vars.update(dict(
                aws_image=os_spec['image'],
                aws_region=env.aws_region,
                aws_ami_id=getattr(env, 'aws_ami_id', None) or None,
            ))
        # Azure case
        if env.cloud == 'azure':
            self.terraform_vars.update(dict(
                azure_region=env.azure_region,
                azure_publisher=os_spec['publisher'],
                azure_offer=os_spec['offer'],
                azure_sku=os_spec['sku']
            ))
        # GCloud case
        if env.cloud == 'gcloud':
            self.terraform_vars.update(dict(
                gcloud_image=os_spec['image'],
                gcloud_region=env.gcloud_region,
                gcloud_credentials=env.gcloud_credentials.name,
                gcloud_project_id=env.gcloud_project_id
            ))

        # Postgres servers terraform_vars
        self.terraform_vars.update(dict(
            postgres_server=dict(
                count=ra['pg_count'],
                instance_type=pg_spec['instance_type'],
                volume=pg_spec['volume'],
                additional_volumes=pg_spec['additional_volumes']
            )
        ))

        # PEM server terraform_vars
        pem_server_spec = env.cloud_spec['pem_server']
        self.terraform_vars.update(dict(
            pem_server=dict(
                count=1 if ra['pem_server'] else 0,
                instance_type=pem_server_spec['instance_type'],
                volume=pem_server_spec['volume']
            )
        ))

        # Barman server terraform_vars
        barman_server_spec = env.cloud_spec['barman_server']
        self.terraform_vars.update(dict(
            barman_server=dict(
                count=1 if ra['barman_server'] else 0,
                instance_type=barman_server_spec['instance_type'],
                volume=barman_server_spec['volume'],
                additional_volumes=barman_server_spec['additional_volumes']
            )
        ))

        # Pooler servers terraform_vars
        pooler_server_spec = env.cloud_spec['pooler_server']
        self.terraform_vars.update(dict(
            pooler_server=dict(
                count=ra['pooler_count'],
                instance_type=pooler_server_spec['instance_type'],
                volume=pooler_server_spec['volume']
            )
        ))

    def _build_ansible_vars(self, env):
        # Fetch EDB repo. username and password
        r = re.compile(r"^([^:]+):(.+)$")
        m = r.search(env.edb_credentials)
        edb_repo_username = m.group(1)
        edb_repo_password = m.group(2)

        os_spec = env.cloud_spec['available_os'][env.operating_system]
        pg_spec = env.cloud_spec['postgres_server']

        self.ansible_vars = dict(
            reference_architecture=env.reference_architecture,
            cluster_name=self.name,
            pg_type=env.postgres_type,
            pg_version=env.postgres_version,
            repo_username=edb_repo_username,
            repo_password=edb_repo_password,
            ssh_user=os_spec['ssh_user'],
            ssh_priv_key=self.custom_ssh_keys[os_spec['ssh_user']]['ssh_priv_key'],
            email_id=env.email_id,
            route53_access_key=env.route53_access_key,
            route53_secret=env.route53_secret,
            project=self.name,
            public_key=self.custom_ssh_keys[self.name]['ssh_pub_key']
        )
        # Add configuration for pg_data and pg_wal accordingly to the number
        # of additional volumes
        if pg_spec['additional_volumes']['count'] > 0:
            self.ansible_vars.update(dict(pg_data='/pgdata/pg_data'))
        if pg_spec['additional_volumes']['count'] > 1:
            self.ansible_vars.update(dict(pg_wal='/pgwal/pg_wal'))

    def show_logs(self, tail):
        if not os.path.exists(self.log_file):
            raise ProjectError("Log file %s not found" % self.log_file)

        if not tail:
            # Read the whole file and write its content to stdout
            with open(self.log_file, "r") as f:
                for l in f.readlines():
                    sys.stdout.write(l)
        else:
            with open(self.log_file, "r") as f:
                # Go to the end of the file
                f.seek(0, 2)
                while True:
                    l = f.readline()
                    if not l:
                        time.sleep(0.1)
                        continue
                    sys.stdout.write(l)

    def remove(self):
        terraform = TerraformCli(
            self.project_path, self.terraform_plugin_cache_path,
            bin_path=self.cloud_tools_bin_path
        )
        # Prevent project deletion if some cloud resources are still present
        # for this project.
        if terraform.count_resources() > 0:
            raise ProjectError(
                "Some cloud resources seem to be still present for this "
                "project, please destroy them with the 'destroy' sub-command"
            )

        if os.path.exists(self.log_file):
            with AM("Removing log file %s" % self.log_file):
                os.unlink(self.log_file)
        with AM("Removing project directory %s" % self.project_path):
            shutil.rmtree(self.project_path)

    def show_configuration(self):
        self._load_terraform_vars()
        self._load_ansible_vars()

        try:
            json_output = json.dumps(
                dict(
                    ansible=self.ansible_vars,
                    terraform=self.terraform_vars
                ),
                indent=2
            )
        except Exception as e:
            msg = "Unable to convert the configuration to JSON"
            logging.error(msg)
            logging.error(str(e))
            raise ProjectError(msg)

        sys.stdout.write(json_output)
        sys.stdout.flush()

    def provision(self):
        terraform = TerraformCli(
            self.project_path, self.terraform_plugin_cache_path,
            bin_path=self.cloud_tools_bin_path
        )

        self.update_state('terraform', 'INITIALIZATING')
        with AM("Terraform project initialization"):
            terraform.init()
        self.update_state('terraform', 'INITIALIZATED')

        self.update_state('terraform', 'PROVISIONING')
        with AM("Applying cloud resources creation"):
            terraform.apply(self.terraform_vars_file)
        self.update_state('terraform', 'PROVISIONED')

        # Checking instance availability
        cloud_cli = CloudCli(self.cloud, bin_path=self.cloud_tools_bin_path)
        self._load_terraform_vars()

        # AWS case
        if self.cloud == 'aws':
            with AM(
                "Checking instances availability in region %s"
                % self.terraform_vars['aws_region']
            ):
                cloud_cli.cli.check_instances_availability(
                    self.terraform_vars['aws_region']
                )
        # Azure case
        if self.cloud == 'azure':
            with AM("Checking instances availability"):
                cloud_cli.cli.check_instances_availability(self.name)
        # GCloud case
        if self.cloud == 'gcloud':
            with AM(
                "Checking instances availability in region %s"
                % self.terraform_vars['gcloud_region']
            ):
                cloud_cli.cli.check_instances_availability(
                    self.name,
                    self.terraform_vars['gcloud_region'],
                    # Total number of nodes
                    (self.terraform_vars['postgres_server']['count']
                     + self.terraform_vars['barman_server']['count']
                     + self.terraform_vars['pem_server']['count']
                     + self.terraform_vars['pooler_server']['count'])
                )


        with AM("SSH configuration"):
            terraform.exec_add_host_sh()

    def destroy(self):
        terraform = TerraformCli(
            self.project_path, self.terraform_plugin_cache_path,
            bin_path=self.cloud_tools_bin_path
        )

        self.update_state('terraform', 'DESTROYING')
        with AM("Destroying cloud resources"):
            terraform.destroy(self.terraform_vars_file)
        self.update_state('terraform', 'DESTROYED')
        self.update_state('ansible', 'UNKNOWN')

    def deploy(self, no_install_collection):
        inventory_data = None
        ansible = AnsibleCli(
                    self.project_path,
                    bin_path = self.cloud_tools_bin_path
                  )

        # Load ansible vars
        self._load_ansible_vars()

        if not no_install_collection:
            with AM(
                "Installing Ansible collection %s"
                % self.ansible_collection_name
            ):
                ansible.install_collection(self.ansible_collection_name)
            with AM(
                "Installing AWS collection %s"
                % self.aws_collection_name
            ):
                ansible.install_collection(self.aws_collection_name)

        # Building extra vars to pass to ansible because it's not safe to pass
        # the content of ansible_vars as it.
        extra_vars=dict(
            pg_type=self.ansible_vars['pg_type'],
            pg_version=self.ansible_vars['pg_version'],
            repo_username=self.ansible_vars['repo_username'],
            repo_password=self.ansible_vars['repo_password'],
            pass_dir=os.path.join(self.project_path, '.edbpass'),
            email_id=self.ansible_vars['email_id'],
            route53_access_key=self.ansible_vars['route53_access_key'],
            route53_secret=self.ansible_vars['route53_secret'],
            project=self.ansible_vars['project'],
            public_key=self.ansible_vars['public_key']
        )
        if self.ansible_vars.get('pg_data'):
            extra_vars.update(dict(
                pg_data=self.ansible_vars['pg_data']
            ))
        if self.ansible_vars.get('pg_wal'):
            extra_vars.update(dict(
                pg_wal=self.ansible_vars['pg_wal']
            ))

        self.update_state('ansible', 'DEPLOYING')
        with AM("Deploying components with Ansible"):
            ansible.run_playbook(
                self.ansible_vars['ssh_user'],
                self.ansible_vars['ssh_priv_key'],
                self.ansible_inventory,
                self.ansible_playbook,
                json.dumps(extra_vars)
            )
        self.update_state('ansible', 'DEPLOYED')

        with AM("Extracting data from the inventory file"):
            inventory_data = ansible.list_inventory(self.ansible_inventory)

        # Display inventory informations
        self.display_inventory(inventory_data)

    def display_details(self):
        try:
            states = self.load_states()
        except Exception as e:
            states={}
        status = states.get('ansible', 'UNKNOWN')
        if status in ['DEPLOYED', 'DEPLOYING']:
            if status == 'DEPLOYING':
                print("WARNING: project is in deploying state")
            inventory_data = None
            ansible = AnsibleCli(
                self.project_path,
                bin_path = self.cloud_tools_bin_path
            )
            with AM("Extracting data from the inventory file"):
                inventory_data = ansible.list_inventory(self.ansible_inventory)
            self.display_inventory(inventory_data)

    def display_passwords(self):
        try:
            states = self.load_states()
        except Exception as e:
            states={}
        status = states.get('ansible', 'UNKNOWN')
        if status in ['DEPLOYED', 'DEPLOYING']:
            if status == 'DEPLOYING':
                print("WARNING: project is in deploying state")
            pass_dir=os.path.join(self.project_path, '.edbpass')
            rows = []
            for pass_file in os.listdir(pass_dir):
                if pass_file.endswith("_pass"):
                    user_name = pass_file.replace('_pass','')
                    with open(
                        os.path.join(
                            pass_dir,
                            pass_file
                        )
                    ) as f:
                        user_password = f.read()
                    rows.append([
                        user_name,
                        user_password.replace('\n','')
                    ])
            Project.display_table(
                ['Username','Password'],
                rows
            )

    def display_inventory(self, inventory_data):
        if not self.ansible_vars:
            self._load_ansible_vars()

        def _p(s):
            sys.stdout.write(s)

        sys.stdout.flush()
        _p("\n")

        # Display PEM server informations
        if 'pemserver' in inventory_data['all']['children']:
            if self.ansible_vars['pg_type'] == 'epas':
                pem_user = 'enterprisedb'
            else:
                pem_user = 'postgres'
            pem_name = inventory_data['pemserver']['hosts'][0]
            pem_hostvars = inventory_data['_meta']['hostvars'][pem_name]

            # In PoT PEM server is client server
            client_login_ip = pem_hostvars['ansible_host']
            client_login_user = self.name
            client_private_ip = pem_hostvars['private_ip']
            se_login_user = self.ansible_vars['ssh_user']

            with open(
                os.path.join(
                    self.project_path, '.edbpass', '%s_pass' % pem_user
                )
            ) as f:
                pem_password = f.read()

            _p(
                "PEM Server: https://%spem.edbpov.io:8443/pem\n"
                % self.name
            )
            _p("PEM User: %s\n" % pem_user)
            _p("PEM Password: %s\n" % pem_password)

        # Build the nodes table
        rows = []
        for name, vars in inventory_data['_meta']['hostvars'].items():
            if name != pem_name:
                rows.append([
                    name,
                    vars['ansible_host'],
                    self.ansible_vars['ssh_user'],
                    vars['private_ip'],
                    self.name
                ])
        rows.append(['client',
                      client_login_ip,
                      se_login_user,
                      client_private_ip,
                      self.name])

        Project.display_table(
            ['Name', 'Login IP Address', 'SE Login User', 'Internal IP Address', 'Login User'],
            rows
        )

    @staticmethod
    def list(cloud):
        projects_path = os.path.join(Project.projects_root_path, cloud)
        rows = []
        try:
            for project_name in os.listdir(projects_path):
                project_path = os.path.join(projects_path, project_name)
                if not os.path.isdir(project_path):
                    continue

                project = Project(cloud, project_name)

                terraform_resource_count = 0
                terraform = TerraformCli(
                    project.project_path,
                    project.terraform_plugin_cache_path,
                    bin_path=project.cloud_tools_bin_path
                )
                terraform_resource_count = terraform.count_resources()

                try:
                    states = project.load_states()
                except Exception as e:
                    states={}

                rows.append([
                    project.name,
                    project.project_path,
                    states.get('terraform', 'UNKNOWN'),
                    str(terraform_resource_count),
                    states.get('ansible', 'UNKNOWN')
                ])

            Project.display_table(
                ["Name", "Path", "Machines", "Resources", "Components"],
                rows
            )

        except OSError as e:
            msg = "Unable to list projects in %s" % projects_path
            logging.error(msg)
            logging.exception(str(e))
            raise ProjectErro(msg)

    @staticmethod
    def display_table(headers, rows):
        def _p(s):
            sys.stdout.write(s)

        # Calculate max lengths
        max_lengths = [0 for i in range(len(headers))]
        for i in range(len(headers)):
            if len(headers[i]) > max_lengths[i]:
                max_lengths[i] = len(headers[i])
        for row in rows:
            for i in range(len(headers)):
                if len(row[i]) > max_lengths[i]:
                    max_lengths[i] = len(row[i])

        _p("\n")
        # Display headers
        for i in range(len(headers)):
            _p(headers[i].center(max_lengths[i] + 4))

        _p("\n")
        _p("=" * (sum(max_lengths) + 4 * len(max_lengths)))
        _p("\n")
        # Display rows
        for row in rows:
            for i in range(len(headers)):
                _p(row[i].rjust(max_lengths[i] + 4))
            _p("\n")
        _p("\n")
        sys.stdout.flush()

    @staticmethod
    def show_specs(cloud):
        sys.stdout.write(json.dumps(default_spec(cloud), indent=2))
        sys.stdout.flush()

    def ssh_key_gen(self, ssh_user, is_customer_key=False):
        if not is_customer_key:
           project_ssh_priv_key = os.path.join(self.project_path, "%s_%s_key" % (ssh_user, self.name))
           project_ssh_pub_key = os.path.join(self.project_path, "%s_%s_key.pub" % (ssh_user, self.name))
        else:
           project_ssh_priv_key = os.path.join(self.project_path, "%s_key" % ssh_user)
           project_ssh_pub_key = os.path.join(self.project_path, "%s_key.pub" % ssh_user)

        try:
            output = exec_shell([
                "ssh-keygen",
                "-q",
                "-t rsa",
                "-f %s" % project_ssh_priv_key,
                "-C \"\" -N \"\"",
            ])
            result = output.decode("utf-8")
            logging.debug("Command output:")
            for l in result.split("\n"):
                logging.debug(l)
        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise ProjectError(
                "Failed to execute the following command, please check the "
                "logs for details: %s" % e.cmd
            )
        ext_project_ssh_priv_key = project_ssh_priv_key + '.pem'
        os.rename(project_ssh_priv_key, ext_project_ssh_priv_key)
        self.custom_ssh_keys[ssh_user] = dict(ssh_pub_key=project_ssh_pub_key,
                                              ssh_priv_key=ext_project_ssh_priv_key)



from . import SpecValidator

AzureSpec = {
    'available_os': {
        'CentOS7': {
            'publisher': SpecValidator(type='string', default="OpenLogic"),
            'offer': SpecValidator(type='string', default="CentOS"),
            'sku': SpecValidator(type='string', default="7.7"),
            'ssh_user': SpecValidator(type='string', default='edbadm')
        },
        'CentOS8': {
            'publisher': SpecValidator(type='string', default="OpenLogic"),
            'offer': SpecValidator(type='string', default="CentOS"),
            'sku': SpecValidator(type='string', default="8_1"),
            'ssh_user': SpecValidator(type='string', default='edbadm')
        },
        'RedHat7': {
            'publisher': SpecValidator(type='string', default="RedHat"),
            'offer': SpecValidator(type='string', default="RHEL"),
            'sku': SpecValidator(type='string', default="7.8"),
            'ssh_user': SpecValidator(type='string', default='edbadm')
        },
        'RedHat8': {
            'publisher': SpecValidator(type='string', default="RedHat"),
            'offer': SpecValidator(type='string', default="RHEL"),
            'sku': SpecValidator(type='string', default="8.2"),
            'ssh_user': SpecValidator(type='string', default='edbadm')
        }
    },
    'postgres_server': {
        'instance_type': SpecValidator(
            type='choice',
            choices=[
                'Standard_A1_v2', 'Standard_A2_v2', 'Standard_A4_v2',
                'Standard_A8_v2', 'Standard_A2m_v2', 'Standard_A4m_v2',
                'Standard_A8m_v2'
            ],
            default='Standard_A4_v2'
        ),
        'volume': {
            'storage_account_type': SpecValidator(
                type='choice',
                choices=['Premium_LRS', 'StandardSSD_LRS', 'Standard_LRS',
                         'UltraSSD_LRS'],
                default='Standard_LRS'
            )
        },
        'additional_volumes': {
            'count': SpecValidator(
                type='integer',
                min=0,
                max=5,
                default=2
            ),
            'storage_account_type': SpecValidator(
                type='choice',
                choices=['Premium_LRS', 'StandardSSD_LRS', 'Standard_LRS',
                         'UltraSSD_LRS'],
                default='StandardSSD_LRS'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=100
            )
        }
    },
    'pem_server': {
        'instance_type': SpecValidator(
            type='choice',
            choices=[
                'Standard_A1_v2', 'Standard_A2_v2', 'Standard_A4_v2',
                'Standard_A8_v2', 'Standard_A2m_v2', 'Standard_A4m_v2',
                'Standard_A8m_v2'
            ],
            default='Standard_A2_v2'
        ),
        'volume': {
            'storage_account_type': SpecValidator(
                type='choice',
                choices=['Premium_LRS', 'StandardSSD_LRS', 'Standard_LRS',
                         'UltraSSD_LRS'],
                default='Standard_LRS'
            )
        }
    },
    'pooler_server': {
        'instance_type': SpecValidator(
            type='choice',
            choices=[
                'Standard_A1_v2', 'Standard_A2_v2', 'Standard_A4_v2',
                'Standard_A8_v2', 'Standard_A2m_v2', 'Standard_A4m_v2',
                'Standard_A8m_v2'
            ],
            default='Standard_A2_v2'
        ),
        'volume': {
            'storage_account_type': SpecValidator(
                type='choice',
                choices=['Premium_LRS', 'StandardSSD_LRS', 'Standard_LRS',
                         'UltraSSD_LRS'],
                default='Standard_LRS'
            )
        }
    },
    'barman_server': {
        'instance_type': SpecValidator(
            type='choice',
            choices=[
                'Standard_A1_v2', 'Standard_A2_v2', 'Standard_A4_v2',
                'Standard_A8_v2', 'Standard_A2m_v2', 'Standard_A4m_v2',
                'Standard_A8m_v2'
            ],
            default='Standard_A2_v2'
        ),
        'volume': {
            'storage_account_type': SpecValidator(
                type='choice',
                choices=['Premium_LRS', 'StandardSSD_LRS', 'Standard_LRS',
                         'UltraSSD_LRS'],
                default='Standard_LRS'
            )
        },
        'additional_volumes': {
            'count': SpecValidator(
                type='integer',
                min=0,
                max=1,
                default=1
            ),
            'storage_account_type': SpecValidator(
                type='choice',
                choices=['Premium_LRS', 'StandardSSD_LRS', 'Standard_LRS',
                         'UltraSSD_LRS'],
                default='StandardSSD_LRS'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=300
            )
        }
    }
}



from . import SpecValidator

GCloudSpec = {
    'available_os': {
        'CentOS7': {
            'image': SpecValidator(type='string', default="centos-7"),
            'ssh_user': SpecValidator(type='string', default='edbadm')
        },
        'CentOS8': {
            'image': SpecValidator(type='string', default="centos-8"),
            'ssh_user': SpecValidator(type='string', default='edbadm')
        },
        'RedHat7': {
            'image': SpecValidator(type='string', default="rhel-7"),
            'ssh_user': SpecValidator(type='string', default='edbadm')
        },
        'RedHat8': {
            'image': SpecValidator(type='string', default="rhel-8"),
            'ssh_user': SpecValidator(type='string', default='edbadm')
        }
    },
    'postgres_server': {
        'instance_type': SpecValidator(
            type='choice',
            choices=[
                'e2-standard-2', 'e2-standard-4', 'e2-standard-8',
                'e2-standard-16', 'e2-standard-32', 'e2-highmem-2',
                'e2-highmem-4', 'e2-highmem-8', 'e2-highmem-16'
            ],
            default='e2-standard-4'
        ),
        'volume': {
            'type': SpecValidator(
                type='choice',
                choices=['pd-standard', 'pd-ssd'],
                default='pd-standard'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=50
            )
        },
        'additional_volumes': {
            'count': SpecValidator(
                type='integer',
                min=0,
                max=5,
                default=2
            ),
            'type': SpecValidator(
                type='choice',
                choices=['pd-standard', 'pd-ssd'],
                default='pd-ssd'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=100
            )
        }
    },
    'pem_server': {
        'instance_type': SpecValidator(
            type='choice',
            choices=[
                'e2-standard-2', 'e2-standard-4', 'e2-standard-8',
                'e2-standard-16', 'e2-standard-32', 'e2-highmem-2',
                'e2-highmem-4', 'e2-highmem-8', 'e2-highmem-16'
            ],
            default='e2-standard-4'
        ),
        'volume': {
            'type': SpecValidator(
                type='choice',
                choices=['pd-standard', 'pd-ssd'],
                default='pd-standard'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=100
            )
        }
    },
    'pooler_server': {
        'instance_type': SpecValidator(
            type='choice',
            choices=[
                'e2-standard-2', 'e2-standard-4', 'e2-standard-8',
                'e2-standard-16', 'e2-standard-32', 'e2-highmem-2',
                'e2-highmem-4', 'e2-highmem-8', 'e2-highmem-16'
            ],
            default='e2-standard-2'
        ),
        'volume': {
            'type': SpecValidator(
                type='choice',
                choices=['pd-standard', 'pd-ssd'],
                default='pd-standard'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=50
            )
        }
    },
    'barman_server': {
        'instance_type': SpecValidator(
            type='choice',
            choices=[
                'e2-standard-2', 'e2-standard-4', 'e2-standard-8',
                'e2-standard-16', 'e2-standard-32', 'e2-highmem-2',
                'e2-highmem-4', 'e2-highmem-8', 'e2-highmem-16'
            ],
            default='e2-standard-2'
        ),
        'volume': {
            'type': SpecValidator(
                type='choice',
                choices=['pd-standard', 'pd-ssd'],
                default='pd-standard'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=50
            )
        },
        'additional_volumes': {
            'count': SpecValidator(
                type='integer',
                min=0,
                max=1,
                default=1
            ),
            'type': SpecValidator(
                type='choice',
                choices=['pd-standard', 'pd-ssd'],
                default='pd-ssd'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=300
            )
        }
    }
}



import logging
from .project import Project

class CommanderError(Exception):
    pass

class Commander:
    def __init__(self, env):
        self.env = env
        self.project = None
        if getattr(self.env, 'project', False):
            self.project = Project(self.env.cloud, self.env.project)

    def _check_project_exists(self):
        if not self.project.exists():
            msg = "Project %s does not exist" % self.project.name
            logging.error(msg)
            raise CommanderError(msg)

    def execute(self):
        logging.info(
            "Executing command: %s %s", self.env.cloud, self.env.sub_command
        )

        if getattr(self, self.env.sub_command, False):
            getattr(self, self.env.sub_command)()
        else:
            raise CommanderError(
                "Sub-command %s not implemented" % self.env.sub_command
            )

    def configure(self):
        if self.project.exists():
            msg = "Project %s already exists" % self.project.name
            logging.error(msg)
            raise CommanderError(msg)

        # Check 3rd party SW versions
        self.project.check_versions()

        logging.info("Project configuration...")
        self.project.create()
        self.project.configure(self.env)
        logging.info("End of project configuration")

    def logs(self):
        self._check_project_exists()
        logging.info("Fetching logs for project %s", self.project.name)
        self.project.show_logs(self.env.tail)

    def remove(self):
        self._check_project_exists()
        logging.info("Removing project %s", self.project.name)
        self.project.remove()

    def show(self):
        self._check_project_exists()
        logging.info("Showing project %s configuration", self.project.name)
        self.project.show_configuration()

    def provision(self):
        self._check_project_exists()

        # Check 3rd party SW versions
        self.project.check_versions()

        logging.info("Provisioning machines for project %s", self.project.name)
        self.project.provision()

    def destroy(self):
        self._check_project_exists()

        # Check 3rd party SW versions
        self.project.check_versions()

        logging.info("Destroying machines for project %s", self.project.name)
        self.project.destroy()

    def deploy(self):
        self._check_project_exists()

        # Check 3rd party SW versions
        self.project.check_versions()

        logging.info("Deploying components for project %s", self.project.name)
        self.project.deploy(self.env.no_install_collection)

    def display(self):
        self._check_project_exists()

        # Check 3rd party SW versions
        self.project.check_versions()

        logging.info("Desplaying project %s details", self.project.name)
        self.project.display_details()

    def passwords(self):
        self._check_project_exists()
        logging.info("Display project %s passwords details", self.project.name)
        self.project.display_passwords()

    def list(self):
        logging.info("Listing project for cloud %s", self.env.cloud)
        Project.list(self.env.cloud)

    def specs(self):
        logging.info("Showing default specs. for cloud %s", self.env.cloud)
        Project.show_specs(self.env.cloud)



ReferenceArchitectureSpec = {
    'EDB-RA-1': {
        'pg_count': 1,
        'pem_server': True,
        'barman': True,
        'barman_server': True,
        'pooler_count': 0,
        'pooler_type': None,
        'pooler_local': False,
        'efm': False,
        'replication_type': None
    },
    'EDB-RA-2': {
        'pg_count': 3,
        'pem_server': True,
        'barman': False,
        'barman_server': False,
        'pooler_count': 0,
        'pooler_type': None,
        'pooler_local': False,
        'efm': True,
        'replication_type': "asynchronous"
    },
    'EDB-RA-3': {
        'pg_count': 3,
        'pem_server': True,
        'barman': True,
        'barman_server': True,
        'pooler_count': 3,
        'pooler_type': "pgpool2",
        'pooler_local': False,
        'efm': True,
        'replication_type': "synchronous"
    }
}



class SpecValidator:
    def __init__(self, type=None, default=None, choices=[], min=None,
                 max=None):
        self.type = type
        self.default = default
        self.choices = choices
        self.min = min
        self.max = max



import logging
import os
import subprocess

def exec_shell(args, environ=os.environ, cwd=None):
    logging.info("Executing command: %s", ' '.join(args))
    return subprocess.check_output(
        ' '.join(args),
        stderr=subprocess.STDOUT,
        shell=True,
        cwd=cwd,
        env=environ
    )

def exec_shell_live(args, environ=os.environ, cwd=None):
    logging.info("Executing command: %s", ' '.join(args))
    logging.debug("environ=%s", environ)
    process = subprocess.Popen(
        ' '.join(args),
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=cwd,
        env=environ
    )

    rc = 0
    while True:
        output = process.stdout.readline()
        if output:
            logging.info(output.decode("utf-8").strip())
        rc = process.poll()
        if rc is not None:
            break

    return rc



import sys
import logging

class ActionManager:
    def __init__(self, msg):
        self.msg = msg

    def __enter__(self):
        logging.info(self.msg)
        sys.stdout.write("%s ... " % self.msg)
        sys.stdout.flush()

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type == None:
            sys.stdout.write("\033[1m\033[92mok\033[0m\n")
        elif exc_type == KeyboardInterrupt:
            sys.stdout.write("\033[1m\033[93maborted\033[0m\n")
        else:
            sys.stdout.write("\033[1m\033[91mfail\033[0m\n")
        sys.stdout.flush()



import logging
import os
import re
from subprocess import CalledProcessError

from .system import exec_shell_live, exec_shell


class TerraformCliError(Exception):
    pass


class TerraformCli:

    def __init__(self, dir, plugin_cache_dir, bin_path=None):
        self.dir = dir
        self.plugin_cache_dir = plugin_cache_dir
        self.environ = os.environ
        self.environ['TF_PLUGIN_CACHE_DIR'] = self.plugin_cache_dir
        # Terraform supported version interval
        self.min_version = (0, 0, 0)
        self.max_version = (0, 14, 7)
        # Path to look up for executable
        self.bin_path = None
        # Force Terraform binary path if bin_path exists and contains
        # terraform file.
        if bin_path is not None and os.path.exists(bin_path):
            if os.path.exists(os.path.join(bin_path, 'terraform')):
                self.bin_path = bin_path

    def check_version(self):
       """
       Verify terraform version, based on the interval formed by min_version
       and max_version.
       Terraform version is fetched using the command: terraform --version
       """
       # note: we do not raise any TerraformCliError from this function
       # because TerraformCliError are used to trigger stuffs when they are
       # catched. In this case, we do not want trigger anything if something
       # fails.
       try:
           output = exec_shell([
               self.bin("terraform"),
               "--version"
           ])
       except CalledProcessError as e:
           logging.error("Failed to execute the command: %s", e.cmd)
           logging.error("Return code is: %s", e.returncode)
           logging.error("Output: %s", e.output)
           raise Exception(
               "Terraform executable seems to be missing. Please install it "
               "or check your PATH variable"
           )

       version = None
       # Parse command output and extract the version number
       pattern = re.compile(r"^Terraform v([0-9]+)\.([0-9]+)\.([0-9]+)$")
       for line in output.decode("utf-8").split("\n"):
           m = pattern.search(line)
           if m:
               version = (int(m.group(1)), int(m.group(2)), int(m.group(3)))
               break

       if version is None:
           raise Exception("Unable to parse Terraform version")

       logging.info("Terraform version: %s", '.'.join(map(str, version)))

       # Verify if the version fetched is supported
       for i in range(0, 3):
           min = self.min_version[i]
           max = self.max_version[i]

           if version[i] < max:
               # If current digit is below the maximum value, no need to
               # check others digits, we are good
               break

           if version[i] not in list(range(min, max + 1)):
               raise Exception(
                   ("Terraform version %s not supported, must be between %s "
                    "and %s") % (
                       '.'.join(map(str, version)),
                       '.'.join(map(str, self.min_version)),
                       '.'.join(map(str, self.max_version)),
                   )
               )

    def bin(self, binary):
        """
        Return binary's path
        """
        if self.bin_path is not None:
            return os.path.join(self.bin_path, binary)
        else:
            return binary

    def init(self):
        try:
            rc = exec_shell_live(
                [self.bin("terraform"), "init", "-no-color", self.dir],
                environ=self.environ,
                cwd=self.dir
            )
            if rc != 0:
                raise Exception("Return code not 0")
        except Exception as e:
            logging.error("Failed to execute the command")
            logging.error(e)
            raise TerraformCliError(
                "Failed to initialize Terraform project, please check the logs"
                " for details."
            )

    def apply(self, vars_file):
        try:
            rc = exec_shell_live(
                [
                    self.bin("terraform"),
                    "apply",
                    "-auto-approve",
                    "-var-file",
                    vars_file,
                    "-no-color",
                    self.dir
                ],
                environ=self.environ,
                cwd=self.dir
            )
            if rc != 0:
                raise Exception("Return code not 0")
        except Exception as e:
            logging.error("Failed to execute the command")
            logging.error(e)
            raise TerraformCliError(
                "Failed to apply Terraform, please check the logs for details."
            )

    def destroy(self, vars_file):
        try:
            rc = exec_shell_live(
                [
                    self.bin("terraform"),
                    "destroy",
                    "-auto-approve",
                    "-var-file",
                    vars_file,
                    "-no-color",
                    self.dir
                ],
                environ=self.environ,
                cwd=self.dir
            )
            if rc != 0:
                raise Exception("Return code not 0")
        except Exception as e:
            logging.error("Failed to execute the command")
            logging.error(e)
            raise TerraformCliError(
                "Failed to destroy Terraform, please check the logs for "
                "details."
            )

    def exec_add_host_sh(self):
        try:
            rc = exec_shell_live(
                ["./add_host.sh"], environ=self.environ, cwd=self.dir
            )
            if rc != 0:
                raise Exception("Return code not 0")
        except Exception as e:
            logging.error("Failed to execute the command")
            logging.error(e)
            raise TerraformCliError(
                ("Failed to execute script %s, please check the logs for "
                 "details.") % os.path.join(self.dir, "add_host.sh")
            )

    def count_resources(self):
        try:
            # Check if the terraform state file exists
            if not os.path.exists(os.path.join(self.dir, 'terraform.tfstate')):
                return 0

            output = exec_shell(
                [self.bin("terraform"), "state", "list"],
                environ=self.environ,
                cwd=self.dir
            )

            result = output.decode("utf-8").split('\n')
            logging.debug("Command output: %s", result)
            return len(result) - 1
        except CalledProcessError as e:
            # Case when the terraform.tfstate file exists but not yet fully
            # usable by terraform. In this case, we just return 0
            if (e.returncode == 1 and
                    "No state file was found!" in e.output.decode("utf-8")):
                return 0

            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise TerraformCliError(
                "Failed to execute the following command, please check the "
                "logs for details: %s" % e.cmd
            )



from . import SpecValidator

# These are instance types to make available to all architectures.
global_instance_choices = ['t4g.nano', 't3.nano', 't3a.nano', 't2.nano']

AWSSpec = {
    'available_os': {
        'CentOS7': {
            'image': SpecValidator(
                type='string',
                default="CentOS Linux 7 x86_64 HVM EBS*"
            ),
            'ssh_user': SpecValidator(
                type='choice',
                choices=['centos'],
                default='centos'
            )
        },
        'CentOS8': {
            'image': SpecValidator(
                type='string',
                default="CentOS 8*"
            ),
            'ssh_user': SpecValidator(
                type='choice',
                choices=['centos'],
                default='centos'
            )
        },
        'RedHat7': {
            'image': SpecValidator(
                type='string',
                default="RHEL-7.8-x86_64*"
            ),
            'ssh_user': SpecValidator(
                type='choice',
                choices=['ec2-user'],
                default='ec2-user'
            )
        },
        'RedHat8': {
            'image': SpecValidator(
                type='string',
                default="RHEL-8.2-x86_64*"
            ),
            'ssh_user': SpecValidator(
                type='choice',
                choices=['ec2-user'],
                default='ec2-user'
            )
        }
    },
    'postgres_server': {
        'instance_type': SpecValidator(
            type='choice',
            choices=[
                'c5.large', 'c5.xlarge', 'c5.2xlarge', 'c5.4xlarge',
                'c5.9xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge',
                'c5.metal'
            ] + global_instance_choices,
            default='c5.2xlarge'
        ),
        'volume': {
            'type': SpecValidator(
                type='choice',
                choices=['io1', 'io2', 'gp2', 'gp3', 'st1', 'sc1'],
                default='gp2'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=50
            ),
            'iops': SpecValidator(
                type='integer',
                min=100,
                max=64000,
                default=250
            )
        },
        'additional_volumes': {
            'count': SpecValidator(
                type='integer',
                min=0,
                max=5,
                default=2
            ),
            'type': SpecValidator(
                type='choice',
                choices=['io1', 'io2', 'gp2', 'gp3', 'st1', 'sc1'],
                default='gp2'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=50
            ),
            'iops': SpecValidator(
                type='integer',
                min=100,
                max=64000,
                default=250
            ),
            'encrypted': SpecValidator(
                type='choice',
                choices=[True, False],
                default=False
            )
        }
    },
    'pem_server': {
        'instance_type': SpecValidator(
            type='choice',
            choices=[
                'c5.large', 'c5.xlarge', 'c5.2xlarge', 'c5.4xlarge',
                'c5.9xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge',
                'c5.metal'
            ] + global_instance_choices,
            default='c5.xlarge'
        ),
        'volume': {
            'type': SpecValidator(
                type='choice',
                choices=['io1', 'io2', 'gp2', 'gp3', 'st1', 'sc1'],
                default='gp2'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=100
            ),
            'iops': SpecValidator(
                type='integer',
                min=100,
                max=64000,
                default=250
            )
        }
    },
    'pooler_server': {
        'instance_type': SpecValidator(
            type='choice',
            choices=[
                'c5.large', 'c5.xlarge', 'c5.2xlarge', 'c5.4xlarge',
                'c5.9xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge',
                'c5.metal'
            ] + global_instance_choices,
            default='c5.xlarge'
        ),
        'volume': {
            'type': SpecValidator(
                type='choice',
                choices=['io1', 'io2', 'gp2', 'gp3', 'st1', 'sc1'],
                default='gp2'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=30
            ),
            'iops': SpecValidator(
                type='integer',
                min=100,
                max=64000,
                default=250
            )
        }
    },
    'barman_server': {
        'instance_type': SpecValidator(
            type='choice',
            choices=[
                'c5.large', 'c5.xlarge', 'c5.2xlarge', 'c5.4xlarge',
                'c5.9xlarge', 'c5.12xlarge', 'c5.18xlarge', 'c5.24xlarge',
                'c5.metal'
            ] + global_instance_choices,
            default='c5.2xlarge'
        ),
        'volume': {
            'type': SpecValidator(
                type='choice',
                choices=['io1', 'io2', 'gp2', 'gp3', 'st1', 'sc1'],
                default='gp2'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=50
            ),
            'iops': SpecValidator(
                type='integer',
                min=100,
                max=64000,
                default=250
            )
        },
        'additional_volumes': {
            'count': SpecValidator(
                type='integer',
                min=0,
                max=1,
                default=1
            ),
            'type': SpecValidator(
                type='choice',
                choices=['io1', 'io2', 'gp2', 'gp3', 'st1', 'sc1'],
                default='gp2'
            ),
            'size': SpecValidator(
                type='integer',
                min=10,
                max=16000,
                default=300
            ),
            'iops': SpecValidator(
                type='integer',
                min=100,
                max=64000,
                default=250
            ),
            'encrypted': SpecValidator(
                type='choice',
                choices=[True, False],
                default=False
            )
        }
    }
}



from .spec import SpecValidator
from .spec.aws import AWSSpec
from .spec.azure import AzureSpec
from .spec.gcloud import GCloudSpec

class SpecValidatorError(Exception):
    pass

def merge(data, spec, defaults, path = []):
    """
    Recursive function for merging user data and cloud specifications.
    """
    # Initialize result to defaults values
    result = defaults

    # Walk through user data
    for key, value in data.items():
        path.append(key)

        if key not in defaults:
            raise SpecValidatorError(
                "%s: unexpected key %s" % ('.'.join(path[:-1]), key)
            )

        if isinstance(value, dict):
            # Merge recursively the value when it is a dictionary
            result[key] = merge(value, spec[key], defaults[key], path)

        elif isinstance(spec[key], SpecValidator):
            # Do user data validation if the corresponding value from the
            # specification is an instance of SpecValidator.
            result[key] = validate(value, spec[key], path)

        else:
            result[key] = value

        path.remove(key)
    return result

def validate(value, validator, path):
    """
    Value validation based on SpecValidator attributes
    """
    if validator.type == 'choice':
        if value not in validator.choices:
            raise SpecValidatorError(
                "%s: value '%s' not allowed, must be in list %s"
                % ('.'.join(path), value, validator.choices)
            )

    elif validator.type == 'integer':
        value = int(value)
        if value < validator.min or value > validator.max:
            raise SpecValidatorError(
                "%s: invalid value %s, must between %s and %s"
                % ('.'.join(path), value, validator.min, validator.max)
            )

    elif validator.type == 'string':
        value = str(value)

    return value

def default(element):
    """
    Recursive function for building specification dictionary based on the
    default values.
    """
    result = {}

    for key, value in element.items():
        if isinstance(value, dict):
            result[key] = default(value)

        elif isinstance(value, SpecValidator):
            result[key] = value.default

        else:
            result[key] = value

    return result

def default_spec(cloud):
    """
    Wrapper for the default function
    """
    if cloud == 'aws':
        return default(AWSSpec)
    elif cloud == 'azure':
        return default(AzureSpec)
    elif cloud == 'gcloud':
        return default(GCloudSpec)
    else:
        return {}

def merge_user_spec(cloud, user_spec):
    """
    Wrapper function for merging user specs. and default specs.
    """
    if cloud == 'aws':
        cloud_spec = AWSSpec
    elif cloud == 'azure':
        cloud_spec = AzureSpec
    elif cloud == 'gcloud':
        cloud_spec = GCloudSpec

    defaults = default(cloud_spec)

    return merge(user_spec, cloud_spec, defaults)



import json
import os
import re
import logging
from subprocess import CalledProcessError

from .system import exec_shell_live, exec_shell


class AnsibleCliError(Exception):
    pass


class AnsibleCli:

    def __init__(self, dir, bin_path=None):
        self.dir = dir
        # Ansible supported versions interval
        self.min_version = (0, 0, 0)
        self.max_version = (2, 10, 7)
        # Path to look up for executable
        self.bin_path = None
        # Force Ansible binary path if bin_path exists and contains
        # ansible file.
        if bin_path is not None and os.path.exists(bin_path):
            if os.path.exists(os.path.join(bin_path, 'ansible')):
                self.bin_path = bin_path

    def check_version(self):
        """
        Verify ansible version, based on the interval formed by min_version and
        max_version.
        Ansible version is fetched using the command: ansible --version
        """
        # note: we do not raise any AnsibleCliError from this function because
        # AnsibleCliError are used to trigger stuffs when they are catched. In
        # this case, we do not want trigger anything if something fails.
        try:
            output = exec_shell([
                self.bin("ansible"),
                "--version"
            ])
        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise Exception(
                "Ansible executable seems to be missing. Please install it or "
                "check your PATH variable"
            )

        version = None
        # Parse command output and extract the version number
        pattern = re.compile(r"^ansible ([0-9]+)\.([0-9]+)\.([0-9]+)$")
        for line in output.decode("utf-8").split("\n"):
            m = pattern.search(line)
            if m:
                version = (int(m.group(1)), int(m.group(2)), int(m.group(3)))
                break

        if version is None:
            raise Exception("Unable to parse Ansible version")

        logging.info("Ansible version: %s", '.'.join(map(str, version)))

        # Verify if the version fetched is supported
        for i in range(0, 3):
            min = self.min_version[i]
            max = self.max_version[i]

            if version[i] < max:
                # If current digit is below the maximum value, no need to
                # check others digits, we are good
                break

            if version[i] not in list(range(min, max + 1)):
                raise Exception(
                    ("Ansible version %s not supported, must be between %s and"
                     " %s") % (
                        '.'.join(map(str, version)),
                        '.'.join(map(str, self.min_version)),
                        '.'.join(map(str, self.max_version)),
                    )
                )

    def bin(self, binary):
        """
        Return binary's path
        """
        if self.bin_path is not None:
            return os.path.join(self.bin_path, binary)
        else:
            return binary

    def install_collection(self, collection_name, version=None):
        if version:
            collection_name += ":%s" % version
        try:
            output = exec_shell([
                self.bin("ansible-galaxy"),
                "collection",
                "install",
                "-f",
                collection_name
            ])
            result = output.decode("utf-8")
            logging.debug("Command output:")
            for l in result.split("\n"):
                logging.debug(l)
        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise AnsibleCliError(
                "Failed to execute the following command, please check the "
                "logs for details: %s" % e.cmd
            )

    def run_playbook(
        self, ssh_user, ssh_priv_key, inventory, playbook, extra_vars
    ):
        try:
            rc = exec_shell_live(
                [
                    self.bin("ansible-playbook"),
                    playbook,
                    "--ssh-common-args='-o StrictHostKeyChecking=no'",
                    "-i", inventory,
                    "-u", ssh_user,
                    "--private-key", ssh_priv_key,
                    "-e", "'%s'" % extra_vars
                ],
                cwd=self.dir
            )
            if rc != 0:
                raise Exception("Return code not 0")
        except Exception as e:
            logging.error("Failed to execute the command")
            logging.error(e)
            raise AnsibleCliError(
                "Failed to execute Ansible playbook, please check the logs for"
                " details."
            )

    def list_inventory(self, inventory):
        try:
            output = exec_shell([
                self.bin("ansible-inventory"),
                "--list",
                "-i", inventory
            ])
            result = json.loads(output.decode("utf-8"))
            logging.debug("Command output: %s", result)
            return result 
        except ValueError:
            # JSON decoding error
            logging.error("Failed to decode JSON data")
            logging.error("Output: %s", output.decode("utf-8"))
            raise AnsibleCliError(
                "Failed to decode JSON data, please check the logs for details"
            )
        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise AnsibleCliError(
                "Failed to execute the following command, please check the "
                "logs for details: %s" % e.cmd
            )



import os.path
import sys

from setuptools import setup
from textwrap import dedent

def get_version():
    cur_dir = os.path.dirname(__file__)
    init_path = os.path.join(cur_dir, "edbpotdeploy", "__init__.py")

    with open(init_path) as f:
        for line in f:
            if line.startswith("__version__"):
                return line.split('"')[1]
    raise Exception("Version information not found in %s" % init_path)

def get_long_description():
    cur_dir = os.path.dirname(__file__)
    with open(os.path.join(cur_dir, "README.md")) as f:
        return f.read()

setup(
    name="edb-pot",
    version=get_version(),
    author="EDB",
    author_email="edb-devops@enterprisedb.com",
    scripts=["edb-pot"],
    packages=["edbpotdeploy", "edbpotdeploy.spec"],
    url="https://github.com/EnterpriseDB/postgres-deployment/",
    license="BSD",
    description=dedent("""
Postgres Deployment Scripts are an easy way to deploy PostgreSQL, EDB Postgres
Advanced Server, and EDB Tools in the Cloud.
    """),
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Database",
    ],
    keywords="postgresql edb epas cli deploy cloud aws azure gcloud",
    python_requires=">=2.7",
    install_requires=["argcomplete"],
    extras_require={},
    data_files=[
        (
            'share/edb-pot/scripts',
            [
                'scripts/install_requirements_linux_x64.sh',
                'scripts/install_requirements_darwin_x64.sh',
            ]
        )
    ],
    package_data={
        'edbpotdeploy': [
            'data/ansible/*.yml',
            'data/ansible/*/*/*/*.yml',
            'data/ansible/*/*/*/*.template',
            'data/terraform/*/*.tf.template',
            'data/terraform/*/*.tf',
            'data/terraform/*/*/*.tf',
            'data/terraform/*/*/*.sh',
            'data/terraform/*/*/*/*.tf',
            'data/terraform/*/*/*/*.sh',
            'data/terraform/*/*/*/*/*.tf',
            'data/terraform/*/*/*/*/*.sh',
        ]
    }
)



import logging
import json
import os
import re
import time
from subprocess import CalledProcessError

from .system import exec_shell


class CloudCliError(Exception):
    pass


class AWSCli:
    def __init__(self, bin_path=None):
        # aws CLI supported versions interval
        self.min_version = (0, 0, 0)
        self.max_version = (1, 19, 18)
        # Path to look up for executable
        self.bin_path = None
        # Force aws CLI binary path if bin_path exists and contains
        # aws file.
        if bin_path is not None and os.path.exists(bin_path):
            if os.path.exists(os.path.join(bin_path, 'aws')):
                self.bin_path = bin_path
        pass

    def check_version(self):
         """
         Verify aws CLI version, based on the interval formed by min_version and
         max_version.
         aws CLI version is fetched using the command: aws --version
         """
         try:
             output = exec_shell([self.bin("aws"), "--version"])
         except CalledProcessError as e:
             logging.error("Failed to execute the command: %s", e.cmd)
             logging.error("Return code is: %s", e.returncode)
             logging.error("Output: %s", e.output)
             raise Exception(
                 "aws CLI executable seems to be missing. Please install it or "
                 "check your PATH variable"
             )

         version = None
         # Parse command output and extract the version number
         pattern = re.compile(r"^aws-cli\/([0-9]+)\.([0-9]+)\.([0-9]+) ")
         for line in output.decode("utf-8").split("\n"):
             m = pattern.search(line)
             if m:
                 version = (int(m.group(1)), int(m.group(2)), int(m.group(3)))
                 break

         if version is None:
             raise Exception("Unable to parse aws CLI version")

         logging.info("aws CLI version: %s", '.'.join(map(str, version)))

         # Verify if the version fetched is supported
         for i in range(0, 3):
             min = self.min_version[i]
             max = self.max_version[i]

             if version[i] < max:
                 # If current digit is below the maximum value, no need to
                 # check others digits, we are good
                 break

             if version[i] not in list(range(min, max + 1)):
                 raise Exception(
                     ("aws CLI version %s not supported, must be between %s and"
                      " %s") % (
                         '.'.join(map(str, version)),
                         '.'.join(map(str, self.min_version)),
                         '.'.join(map(str, self.max_version)),
                     )
                 )

    def bin(self, binary):
        """
        Return binary's path
        """
        if self.bin_path is not None:
            return os.path.join(self.bin_path, binary)
        else:
            return binary

    def check_instance_type_availability(self, instance_type, region):
        try:
            output = exec_shell([
                self.bin("aws"),
                "ec2",
                "describe-instance-type-offerings",
                "--location-type availability-zone",
                "--filters Name=instance-type,Values=%s" % instance_type,
                "--region %s" % region,
                "--output json"
            ])
            result = json.loads(output.decode("utf-8"))
            logging.debug("Command output: %s", result)
            if len(result["InstanceTypeOfferings"]) == 0:
                raise CloudCliError(
                    "Instance type %s not available in region %s"
                    % (instance_type, region)
                )
        except ValueError:
            # JSON decoding error
            logging.error("Failed to decode JSON data")
            logging.error("Output: %s", output.decode("utf-8"))
            raise CloudCliError(
                "Failed to decode JSON data, please check the logs for details"
            )
        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise CloudCliError(
                "Failed to execute the following command, please check the "
                "logs for details: %s" % e.cmd
            )

    def get_image_id(self, image, region):
        try:
            output = exec_shell([
                self.bin("aws"),
                "ec2",
                "describe-images",
                "--filters Name=name,Values=\"%s\"" % image,
                "--query 'sort_by(Images, &Name)[-1]'",
                "--region %s" % region,
                "--output json"
            ])
            result = json.loads(output.decode("utf-8"))
            logging.debug("Command output: %s", result)

            if result.get('State') == 'available':
                return result.get('ImageId')

        except ValueError:
            # JSON decoding error
            logging.error("Failed to decode JSON data")
            logging.error("Output: %s", output.decode("utf-8"))
            raise CloudCliError(
                "Failed to decode JSON data, please check the logs for details"
            )
        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise CloudCliError(
                "Failed to execute the following command, please check the "
                "logs for details: %s" % e.cmd
            )

    def check_instances_availability(self, region):
        try:
            output = exec_shell([
                self.bin("aws"),
                "ec2",
                "wait",
                "instance-status-ok",
                "--region %s" % region
            ])
            logging.debug("Command output: %s", output.decode("utf-8"))

        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise CloudCliError(
                "Failed to execute the following command, please check the "
                "logs for details: %s" % e.cmd
            )


class AzureCli:
    def __init__(self, bin_path=None):
         # azure CLI supported versions interval
         self.min_version = (0, 0, 0)
         self.max_version = (2, 20, 0)
         # Path to look up for executable
         self.bin_path = None
         # Force azure CLI binary path if bin_path exists and contains
         # az file.
         if bin_path is not None and os.path.exists(bin_path):
             if os.path.exists(os.path.join(bin_path, 'az')):
                 self.bin_path = bin_path

    def check_version(self):
        """
        Verify azure CLI version, based on the interval formed by min_version and
        max_version.
        azure CLI version is fetched using the command: az --version
        """
        try:
            output = exec_shell([self.bin(self.bin("az")), "--version"])
        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise Exception(
                "azure CLI executable seems to be missing. Please install it or "
                "check your PATH variable"
            )

        version = None
        # Parse command output and extract the version number
        pattern = re.compile(r"^azure-cli\s+([0-9]+)\.([0-9]+)\.([0-9]+)$")
        for line in output.decode("utf-8").split("\n"):
            m = pattern.search(line)
            if m:
                version = (int(m.group(1)), int(m.group(2)), int(m.group(3)))
                break

        if version is None:
            raise Exception("Unable to parse azure CLI version")

        logging.info("azure CLI version: %s", '.'.join(map(str, version)))

        # Verify if the version fetched is supported
        for i in range(0, 3):
            min = self.min_version[i]
            max = self.max_version[i]

            if version[i] < max:
                # If current digit is below the maximum value, no need to
                # check others digits, we are good
                break

            if version[i] not in list(range(min, max + 1)):
                raise Exception(
                    ("azure CLI version %s not supported, must be between %s and"
                     " %s") % (
                        '.'.join(map(str, version)),
                        '.'.join(map(str, self.min_version)),
                        '.'.join(map(str, self.max_version)),
                    )
                )

    def bin(self, binary):
        """
        Return binary's path
        """
        if self.bin_path is not None:
            return os.path.join(self.bin_path, binary)
        else:
            return binary

    def check_instance_type_availability(self, instance_type, region):
        try:
            output = exec_shell([
                self.bin("az"),
                "vm",
                "list-sizes",
                "--location %s" % region,
                "--query \"[?name == '%s']\"" % instance_type,
                "--output json"
            ])
            result = json.loads(output.decode("utf-8"))
            logging.debug("Command output: %s", result)
            if len(result) == 0:
                raise CloudCliError(
                    "Instance type %s not available in region %s"
                    % (instance_type, region)
                )
        except ValueError:
            # JSON decoding error
            logging.error("Failed to decode JSON data")
            logging.error("Output: %s", output.decode("utf-8"))
            raise CloudCliError(
                "Failed to decode JSON data, please check the logs for details"
            )
        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise CloudCliError(
                "Failed to execute the following command, please check the "
                "logs for details: %s" % e.cmd
            )

    def check_image_availability(self, publisher, offer, sku, region):
        try:
            output = exec_shell([
                self.bin("az"),
                "vm",
                "image",
                "list",
                "--all",
                "-p \"%s\"" % publisher,
                "-f \"%s\"" % offer,
                "-s \"%s\"" % sku,
                "-l %s" % region,
                "--query",
                "\"[?offer == '%s' && sku =='%s']\"" % (offer, sku),
                "--output json"
            ])
            result = json.loads(output.decode("utf-8"))
            logging.debug("Command output: %s", result)
            if len(result) == 0:
                raise CloudCliError(
                    "Image %s:%s:%s not available in region %s"
                    % (publisher, offer, sku, region)
                )
        except ValueError:
            # JSON decoding error
            logging.error("Failed to decode JSON data")
            logging.error("Output: %s", output.decode("utf-8"))
            raise CloudCliError(
                "Failed to decode JSON data, please check the logs for details"
            )
        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise CloudCliError(
                "Failed to execute the following command, please check the "
                "logs for details: %s" % e.cmd
            )

    def check_instances_availability(self, project_name):
        try:
            output = exec_shell([
                self.bin("az"),
                "vm",
                "wait",
                "--ids",
                "$(%s vm list -g \"%s_edb_resource_group\" --query \"[].id\" -o tsv)"
                % (self.bin("az"), project_name),
                "--created"
            ])
            logging.debug("Command output: %s", output.decode("utf-8"))

        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise CloudCliError(
                "Failed to execute the following command, please check the "
                "logs for details: %s" % e.cmd
            )


class GCloudCli:
    def __init__(self):
        # gcloud CLI supported versions interval
        self.min_version = (0, 0, 0)
        self.max_version = (329, 0, 0)
        # Path to look up for executable
        self.bin_path = None
        # Force gcloud CLI binary path if bin_path exists and contains
        # gcloud file.
        if bin_path is not None and os.path.exists(bin_path):
            if os.path.exists(os.path.join(bin_path, 'gcloud')):
                self.bin_path = bin_path

    def check_version(self):
        """
        Verify gcloud CLI version, based on the interval formed by min_version and
        max_version.
        gcloud CLI version is fetched using the command: gcloud --version
        """
        try:
            output = exec_shell([self.bin(self.bin("gcloud")), "--version"])
        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise Exception(
                "gcloud CLI executable seems to be missing. Please install it or "
                "check your PATH variable"
            )

        version = None
        # Parse command output and extract the version number
        pattern = re.compile(r"^Google Cloud SDK ([0-9]+)\.([0-9]+)\.([0-9]+)")
        for line in output.decode("utf-8").split("\n"):
            m = pattern.search(line)
            if m:
                version = (int(m.group(1)), int(m.group(2)), int(m.group(3)))
                break

        if version is None:
            raise Exception("Unable to parse gcloud CLI version")

        logging.info("gcloud CLI version: %s", '.'.join(map(str, version)))

        # Verify if the version fetched is supported
        for i in range(0, 3):
            min = self.min_version[i]
            max = self.max_version[i]

            if version[i] < max:
                # If current digit is below the maximum value, no need to
                # check others digits, we are good
                break

            if version[i] not in list(range(min, max + 1)):
                raise Exception(
                    ("gcloud CLI version %s not supported, must be between %s and"
                     " %s") % (
                        '.'.join(map(str, version)),
                        '.'.join(map(str, self.min_version)),
                        '.'.join(map(str, self.max_version)),
                    )
                )

    def bin(self, binary):
        """
        Return binary's path
        """
        if self.bin_path is not None:
            return os.path.join(self.bin_path, binary)
        else:
            return binary

    def check_instance_type_availability(self, instance_type, region):
        try:
            output = exec_shell([
                self.bin("gcloud"),
                "compute",
                "machine-types",
                "list",
                "--filter=\"name=%s zone:%s*\"" % (instance_type, region),
                "--format=json"
            ])
            result = json.loads(output.decode("utf-8"))
            logging.debug("Command output: %s", result)
            if len(result) == 0:
                raise CloudCliError(
                    "Instance type %s not available in region %s"
                    % (instance_type, region)
                )
        except ValueError:
            # JSON decoding error
            logging.error("Failed to decode JSON data")
            logging.error("Output: %s", output.decode("utf-8"))
            raise CloudCliError(
                "Failed to decode JSON data, please check the logs for details"
            )
        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise CloudCliError(
                "Failed to execute the following command, please check the "
                "logs for details: %s" % e.cmd
            )

    def check_image_availability(self, image):
        try:
            output = exec_shell([
                self.bin("gcloud"),
                "compute",
                "images",
                "list",
                "--filter=\"family=%s\"" % image,
                "--format=json"
            ])
            result = json.loads(output.decode("utf-8"))
            logging.debug("Command output: %s", result)
            if len(result) == 0 or result[0]['status'] != 'READY':
                raise CloudCliError("Image %s not available" % image)
        except ValueError:
            # JSON decoding error
            logging.error("Failed to decode JSON data")
            logging.error("Output: %s", output.decode("utf-8"))
            raise CloudCliError(
                "Failed to decode JSON data, please check the logs for details"
            )
        except CalledProcessError as e:
            logging.error("Failed to execute the command: %s", e.cmd)
            logging.error("Return code is: %s", e.returncode)
            logging.error("Output: %s", e.output)
            raise CloudCliError(
                "Failed to execute the following command, please check the "
                "logs for details: %s" % e.cmd
            )

    def check_instances_availability(self, project_name, region, node_count):
        try_count = 0
        try_max = 5
        try_nap_time = 2
        while True:
            if try_count >= try_max:
                raise CloudCliError(
                    "Unable to check instances availability after %s trys"
                    % try_count
                )

            try_count += 1

            try:
                output = exec_shell([
                    self.bin("gcloud"),
                    "compute",
                    "instances",
                    "list",
                    "--filter=\"name:%s-* zone ~ %s-[a-z] status=RUNNING\""
                    % (project_name, region),
                    "--format=json"
                ])
                result = json.loads(output.decode("utf-8"))
                logging.debug("Command output: %s", result)

                if (len(result) >= node_count):
                    # Number of ready instances is good, just break the loop
                    break

                time.sleep(try_nap_time)

            except ValueError:
                # JSON decoding error
                logging.error("Failed to decode JSON data")
                logging.error("Output: %s", output.decode("utf-8"))
                raise CloudCliError(
                    "Failed to decode JSON data, please check the logs for "
                    "details"
                )
            except CalledProcessError as e:
                logging.error("Failed to execute the command: %s", e.cmd)
                logging.error("Return code is: %s", e.returncode)
                logging.error("Output: %s", e.output)
                raise CloudCliError(
                    "Failed to execute the following command, please check the"
                    " logs for details: %s" % e.cmd
                )


class CloudCli:

    def __init__(self, cloud, bin_path):
        self.cloud = cloud
        if self.cloud == 'aws':
            self.cli = AWSCli(bin_path)
        elif self.cloud == 'azure':
            self.cli = AzureCli(bin_path)
        elif self.cloud == 'gcloud':
            self.cli = GCloudCli(bin_path)
        else:
            raise Exception("Unknown cloud %s", self.cloud)

    def check_instance_type_availability(self, instance_type, region):
        return self.cli.check_instance_type_availability(instance_type, region)

    def check_version(self):
         self.cli.check_version()



import argcomplete
import argparse
import os
import re
import sys
import textwrap

from . import __version__
from .project import Project

class PgTypeOption:
    choices = ['PG', 'EPAS']
    default = 'EPAS'
    help = textwrap.dedent("""
        Postgres engine type. Allowed values are: PG for PostgreSQL, EPAS for
        EDB Postgres Advanced Server. Default: %(default)s
    """)



# Cloud specific options
class AWSRegionOption:
    choices = ['us-east-1', 'us-east-2', 'us-west-1', 'us-west-2']
    default = 'us-east-1'
    help = textwrap.dedent("""
        AWS region. Allowed values are us-east-1, us-east-2, us-west-1 and
        us-west-2. Default: %(default)s
    """)


class AWSIAMIDOption:
    default = ''
    help = textwrap.dedent("""
        AWS Image ID. Default: %(default)s
    """)


class AzureRegionOption:
    choices = ['centralus', 'eastus', 'eastus2', 'westus', 'westcentralus',
               'westus2', 'northcentralus', 'southcentralus']
    default = 'eastus'
    help = textwrap.dedent("""
        Azure region. Allowed values are centralus, eastus, eastus2, westus,
        westcentralus, westus2, northcentralus and southcentralus.
        Default: %(default)s
    """)


class GCloudRegionOption:
    choices = ['us-central1', 'us-east1', 'us-east4', 'us-west1', 'us-west2',
               'us-west3', 'us-west4']
    default = 'us-east1'
    help = textwrap.dedent("""
        GCloud region. Allowed values are us-central1, us-east1, us-east4,
        us-west1, us-west2, us-west3 and us-west4. Default: %(default)s
    """)


class GCloudCredentialsOption:

    help = textwrap.dedent("""
        GCloud credentials file (JSON) to use. Default: %(default)s
    """)

    @staticmethod
    def default():
        home = os.path.expanduser("~")
        credential_file = os.path.join(home, 'accounts.json')
        if os.path.exists(credential_file):
            return credential_file


def EDBCredentialsType(value):
    p = re.compile(r"^([^:]+):(.+)$")
    if not p.match(value):
        raise argparse.ArgumentTypeError(
            "EDB Credentials does not match \"<username>:<password>\""
        )
    return value


def ProjectType(value):
    p = re.compile(r"^[a-z0-9]{3,12}$")
    if not p.match(value):
        raise argparse.ArgumentTypeError(
            "Project name should only contain lower alphanumeric characters, "
            "length must be between 3 and 12"
        )
    return value


class CLIParser(argparse.ArgumentParser):
    def error(self, message):
        if message == "too few arguments":
            self.print_help()
        sys.stderr.write('error: %s\n' % message)
        sys.exit(2)


# AWS sub-commands and options
def aws_subcommands(aws_subparser):
    aws_configure = aws_subparser.add_parser(
        'configure', help='Project configuration'
    )
    aws_provision = aws_subparser.add_parser(
        'provision', help='Machines provisioning'
    )
    aws_destroy = aws_subparser.add_parser(
        'destroy', help='Machines destruction'
    )
    aws_deploy = aws_subparser.add_parser(
        'deploy', help='Postgres deployment'
    )
    aws_show = aws_subparser.add_parser(
        'show', help='Show configuration'
    )
    aws_display = aws_subparser.add_parser(
        'display', help='Display project details'
    )
    aws_passwords = aws_subparser.add_parser(
        'passwords', help='Display project password'
    )
    aws_list = aws_subparser.add_parser(
        'list', help='List projects'
    )
    aws_specs = aws_subparser.add_parser(
        'specs', help='Show Cloud default specifications'
    )
    aws_logs = aws_subparser.add_parser(
        'logs', help='Show project logs'
    )
    aws_remove = aws_subparser.add_parser(
        'remove', help='Remove project'
    )
    # aws configure sub-command options
    aws_configure.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    aws_configure.add_argument(
        '--route53-access-key',
        dest='route53_access_key',
        required=True,
        type=str,
        metavar='<route53-acccess-key>',
        help="Route53 Access Key"
    )
    aws_configure.add_argument(
        '--route53-secret',
        dest='route53_secret',
        required=True,
        type=str,
        metavar='<route53-secret>',
        help="Route53 Secret"
    )
    aws_configure.add_argument(
        '--email-id',
        dest='email_id',
        required=True,
        type=str,
        metavar='<email-id>',
        help="Email Id"
    )
    aws_configure.add_argument(
        '-u', '--edb-credentials',
        dest='edb_credentials',
        required=True,
        type=EDBCredentialsType,
        metavar='"<username>:<password>"',
        help="EDB Packages repository credentials."
    ).completer = edb_credentials_completer
    aws_configure.add_argument(
        '-t', '--pg-type',
        dest='postgres_type',
        choices=PgTypeOption.choices,
        default=PgTypeOption.default,
        metavar='<postgres-engine-type>',
        help=PgTypeOption.help
    )
    aws_configure.add_argument(
        '-r', '--aws-region',
        dest='aws_region',
        choices=AWSRegionOption.choices,
        default=AWSRegionOption.default,
        metavar='<cloud-region>',
        help=AWSRegionOption.help
    )
    aws_configure.add_argument(
        '-i', '--aws-ami-id',
        dest='aws_ami_id',
        type=str,
        default=AWSIAMIDOption.default,
        metavar='<aws-ami-id>',
        help=AWSIAMIDOption.help
    ).completer = aws_ami_id_completer
    aws_configure.add_argument(
        '-s', '--spec',
        dest='spec_file',
        type=argparse.FileType('r'),
        metavar='<aws-spec-file>',
        help="AWS instances specification file, in JSON."
    )
    # aws logs sub-command options
    aws_logs.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    aws_logs.add_argument(
        '-t', '--tail',
        dest='tail',
        action='store_true',
        help="Do not stop at the end of file."
    )
    # aws remove sub-command options
    aws_remove.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # aws show sub-command options
    aws_show.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # aws display sub-command options
    aws_display.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # aws password sub-command option
    aws_passwords.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # aws provision sub-command options
    aws_provision.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # aws destroy sub-command options
    aws_destroy.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # aws deploy sub-command options
    aws_deploy.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    aws_deploy.add_argument(
        '-n', '--no-install-collection',
        dest='no_install_collection',
        action='store_true',
        help="Do not install the Ansible collection."
    )

# Azure sub-commands and options
def azure_subcommands(azure_subparser):
    azure_configure = azure_subparser.add_parser(
        'configure', help='Project configuration'
    )
    azure_provision = azure_subparser.add_parser(
        'provision', help='Machines provisionning'
    )
    azure_destroy = azure_subparser.add_parser(
        'destroy', help='Machines destruction'
    )
    azure_deploy = azure_subparser.add_parser(
        'deploy', help='Postgres deployment'
    )
    azure_show = azure_subparser.add_parser(
        'show', help='Show configuration'
    )
    azure_display = azure_subparser.add_parser(
        'display', help='Display project details'
    )
    azure_passwords = azure_subparser.add_parser(
        'passwords', help='Display project passwords'
    )
    azure_list = azure_subparser.add_parser(
        'list', help='List projects'
    )
    azure_specs = azure_subparser.add_parser(
        'specs', help='Show Cloud default specifications'
    )
    azure_logs = azure_subparser.add_parser(
        'logs', help='Show project logs'
    )
    azure_remove = azure_subparser.add_parser(
        'remove', help='Remove project'
    )
    # azure configure sub-command options
    azure_configure.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Terraform project name'
    ).completer = project_name_completer
    azure_configure.add_argument(
        '--route53-access-key',
        dest='route53_access_key',
        required=True,
        type=str,
        metavar='<route53-acccess-key>',
        help="Route53 Access Key"
    )
    azure_configure.add_argument(
        '--route53-secret',
        dest='route53_secret',
        required=True,
        type=str,
        metavar='<route53-secret>',
        help="Route53 Secret"
    )
    azure_configure.add_argument(
        '--email-id',
        dest='email_id',
        required=True,
        type=str,
        metavar='<email-id>',
        help="Email Id"
    )
    azure_configure.add_argument(
        '-u', '--edb-credentials',
        dest='edb_credentials',
        required=True,
        type=EDBCredentialsType,
        metavar='"<username>:<password>"',
        help="EDB Packages repository credentials."
    ).completer = edb_credentials_completer
    azure_configure.add_argument(
        '-t', '--pg-type',
        dest='postgres_type',
        choices=PgTypeOption.choices,
        default=PgTypeOption.default,
        metavar='<postgres-engine-type>',
        help=PgTypeOption.help
    )
    azure_configure.add_argument(
        '-s', '--spec',
        dest='spec_file',
        type=argparse.FileType('r'),
        metavar='<azure-spec-file>',
        help="Azure instances specification file, in JSON."
    )
    azure_configure.add_argument(
        '-r', '--azure-region',
        dest='azure_region',
        choices=AzureRegionOption.choices,
        default=AzureRegionOption.default,
        metavar='<cloud-region>',
        help=AzureRegionOption.help
    )
    # azure logs sub-command options
    azure_logs.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = edb_credentials_completer
    azure_logs.add_argument(
        '-t', '--tail',
        dest='tail',
        action='store_true',
        help="Do not stop at the end of file."
    )
    # azure remove sub-command options
    azure_remove.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # azure show sub-command options
    azure_show.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # azure display sub-command options
    azure_display.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # azure password sub-command options
    azure_passwords.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # azure provision sub-command options
    azure_provision.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # azure destroy sub-command options
    azure_destroy.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # azure deploy sub-command options
    azure_deploy.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    azure_deploy.add_argument(
        '-n', '--no-install-collection',
        dest='no_install_collection',
        action='store_true',
        help="Do not install the Ansible collection."
    )

# GCloud sub-commands and options
def gcloud_subcommands(gcloud_subparser):
    gcloud_configure = gcloud_subparser.add_parser(
        'configure', help='Project configuration'
    )
    gcloud_provision = gcloud_subparser.add_parser(
        'provision', help='Machines provisionning'
    )
    gcloud_destroy = gcloud_subparser.add_parser(
        'destroy', help='Machines destruction'
    )
    gcloud_deploy = gcloud_subparser.add_parser(
        'deploy', help='Postgres deployment'
    )
    gcloud_show = gcloud_subparser.add_parser(
        'show', help='Show configuration'
    )
    gcloud_display = gcloud_subparser.add_parser(
        'display', help='Display project details'
    )
    gcloud_passwords = gcloud_subparser.add_parser(
        'passwords', help='Display project passwords'
    )
    gcloud_list = gcloud_subparser.add_parser(
        'list', help='List projects'
    )
    gcloud_specs = gcloud_subparser.add_parser(
        'specs', help='Show Cloud default specifications'
    )
    gcloud_logs = gcloud_subparser.add_parser(
        'logs', help='Show project logs'
    )
    gcloud_remove = gcloud_subparser.add_parser(
        'remove', help='Remove project'
    )
    # gcloud configure sub-command options
    gcloud_configure.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Terraform project name'
    ).completer = edb_credentials_completer
    gcloud_configure.add_argument(
        '--route53-access-key',
        dest='route53_access_key',
        required=True,
        type=str,
        metavar='<route53-acccess-key>',
        help="Route53 Access Key"
    )
    gcloud_configure.add_argument(
        '--route53-secret',
        dest='route53_secret',
        required=True,
        type=str,
        metavar='<route53-secret>',
        help="Route53 Secret"
    )
    gcloud_configure.add_argument(
        '--email-id',
        dest='email_id',
        required=True,
        type=str,
        metavar='<email-id>',
        help="Email Id"
    )
    gcloud_configure.add_argument(
        '-u', '--edb-credentials',
        dest='edb_credentials',
        required=True,
        type=EDBCredentialsType,
        metavar='"<username>:<password>"',
        help="EDB Packages repository credentials."
    ).completer = edb_credentials_completer
    gcloud_configure.add_argument(
        '-t', '--pg-type',
        dest='postgres_type',
        choices=PgTypeOption.choices,
        default=PgTypeOption.default,
        metavar='<postgres-engine-type>',
        help=PgTypeOption.help
    )
    gcloud_configure.add_argument(
        '-r', '--gcloud-region',
        dest='gcloud_region',
        choices=GCloudRegionOption.choices,
        default=GCloudRegionOption.default,
        metavar='<cloud-region>',
        help=GCloudRegionOption.help
    )
    gcloud_configure.add_argument(
        '-s', '--spec',
        dest='spec_file',
        type=argparse.FileType('r'),
        metavar='<gcloud-spec-file>',
        help="GCloud instances specification file, in JSON."
    )
    gcloud_configure.add_argument(
        '-c', '--gcloud-credentials',
        dest='gcloud_credentials',
        type=argparse.FileType('r'),
        default=GCloudCredentialsOption.default(),
        metavar='<gcloud-credentials-json-file>',
        help=GCloudCredentialsOption.help
    )
    gcloud_configure.add_argument(
        '-p', '--gcloud-project-id',
        dest='gcloud_project_id',
        required=True,
        type=str,
        metavar='<gcloud-project-id>',
        help="GCloud project ID"
    ).completer = edb_credentials_completer
    # gcloud logs sub-command options
    gcloud_logs.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    gcloud_logs.add_argument(
        '-t', '--tail',
        dest='tail',
        action='store_true',
        help="Do not stop at the end of file."
    )
    # gcloud remove sub-command options
    gcloud_remove.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # gcloud show sub-command options
    gcloud_show.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # gcloud display sub-command options
    gcloud_display.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # gcloud password sub-command options
    gcloud_passwords.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # gcloud provision sub-command options
    gcloud_provision.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # gcloud destroy sub-command options
    gcloud_destroy.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    # gcloud deploy sub-command options
    gcloud_deploy.add_argument(
        'project', type=ProjectType, metavar='<project-name>',
        help='Project name'
    ).completer = project_name_completer
    gcloud_deploy.add_argument(
        '-n', '--no-install-collection',
        dest='no_install_collection',
        action='store_true',
        help="Do not install the Ansible collection."
    )

# Argcomplete completer
def project_name_completer(prefix, parsed_args, **kwargs):
    cloud = parsed_args.cloud
    sub_command = parsed_args.sub_command
    if sub_command == 'configure':
        return ("PROJECT_NAME",)
    else:
        projects_path = os.path.join(Project.projects_root_path, cloud)
        if not os.path.exists(projects_path):
            return ("PROJECT_NAME",)
        return (pname for pname in os.listdir(projects_path))

def edb_credentials_completer(prefix, parsed_args, **kwargs):
    return ("USERNAME:PASSWORD",)

def aws_ami_id_completer(prefix, parsed_args, **kwargs):
    return ("AWS_AMI_ID",)

def gcloud_project_id_completer(prefix, parsed_args, **kwargs):
    return ("GCLOUD_PROJECT_ID",)

def parse():
    parser = CLIParser(
        description='EDB deployment script for aws, azure and gcloud'
    )
    parser.add_argument(
        '-v', '--version',
        dest='version',
        action='store_true',
        help="show version."
    )
    subparsers = parser.add_subparsers(
        title='Cloud provider', dest='cloud', metavar='<cloud>'
    )

    # Cloud commands
    aws = subparsers.add_parser('aws', help='AWS Cloud')
    azure = subparsers.add_parser('azure', help='Azure Cloud')
    gcloud = subparsers.add_parser('gcloud', help='Google Cloud')

    # Sub-commands
    # AWS
    aws_subparser = aws.add_subparsers(
        title='AWS sub-commands', dest='sub_command', metavar='<sub-command>'
    )
    aws_subcommands(aws_subparser)
    # Azure
    azure_subparser = azure.add_subparsers(
        title='Azure sub-commands', dest='sub_command', metavar='<sub-command>'
    )
    azure_subcommands(azure_subparser)
    # GCloud
    gcloud_subparser = gcloud.add_subparsers(
        title='GCloud sub-commands',
        dest='sub_command',
        metavar='<sub-command>'
    )
    gcloud_subcommands(gcloud_subparser)

    # Autocompletion with argcomplete
    argcomplete.autocomplete(parser)

    # Parse the arguments and options
    env = parser.parse_args()

    # -v / --version
    # Show version and exit
    if env.version:
        print(__version__)
        sys.exit(0)

    # Check if the <cloud> argument is set
    if not getattr(env, 'cloud'):
        parser.print_help()
        sys.stderr.write('error: too few arguments\n')
        sys.exit(2)
    # Check if the <sub-command> argument is set
    if not getattr(env, 'sub_command'):
        if env.cloud == 'aws':
            aws.print_help()
        elif env.cloud == 'azure':
            azure.print_help()
        elif env.cloud == 'gcloud':
            gcloud.print_help()
        sys.stderr.write('error: too few arguments\n')
        sys.exit(2)

    return env
